package xyz.matj1.bp;

import mcdr.objects.utils.RankedSortedCollectionDistHashcode;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import messif.operations.query.KNNQueryOperation;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtils;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import xyz.matj1.bp.Utils.Pair;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.function.BiFunction;

;


/**
 * Tato třída je pro testování motion words tak,
 * že se vyhledávají nejpodobnější pohyby k danému pohybu s určitou délkou,
 * a měří se, jak pravděpodobně jsou mezi těmi nejpodobnějšími pohyby
 * různě dlouhé pohyby ze stejné třídy.
 *
 * @author Matěj Bagar
 */
public class LengthTesting {
	public static void lengthTesting(ObjectMgmt mwMgmt) {
		var categoryObjects = mwMgmt.getCategoryObjects();
		for (var list : categoryObjects.values()) {
			list.sort(Comparator.comparing(Utils::getLength));
		} // seřazení sequencí v seznamu v každé hodnotě té mapy podle délky

		for (var entry : categoryObjects.entrySet()) {
			var category = entry.getKey();
			var sequences = entry.getValue();
			for (var sequence : sequences) {
				var operation = new KNNQueryOperation(sequence, sequences.size() - 1);
				operation.setAnswerCollection(new RankedSortedCollectionDistHashcode());

				mwMgmt.getObjects().forEach(object -> {
					if (!object.getLocatorURI().equals(sequence.getLocatorURI())) {
						operation.addToAnswer(object);
				}});

				var answer =
					Utils.iteratorToStream(operation.getAnswer())
						.map(o -> (LocalAbstractObject)(o.getObject()));

				var answerScore = answer.mapToInt(o ->
					mwMgmt
						.getObjectCategories(o)
						.contains(category)
					? 1 : 0
				).sum();

				var averageAnswerLength = answer
					.mapToInt(Utils::getLength)
					.average()
					.orElseThrow();


			}
		}
	}

	/**
	 * Calculates the length relation and the distance for each pair of sequences in each category.
	 * <!-- TODO: quantisovat rozdíly délek? -->
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @return a map of categories to maps of pairs of the sequences to the results
	 */
	public static Map<Category, Map<Pair<LocalAbstractObject, LocalAbstractObject>, Pair<Double, Double>>> pairwiseCorrelation(
		ObjectMgmt mwMgmt,
		Utils.LengthRelation lengthRelation
	) {
		var categoryObjects = mwMgmt.getCategoryObjects();
		var accumulator = new HashMap<Category, Map<Pair<LocalAbstractObject, LocalAbstractObject>, Pair<Double, Double>>>();

		for (var entry : categoryObjects.entrySet()) {
			var category = entry.getKey();
			var sequences = entry.getValue();
			accumulator.put(category, new HashMap<>());

			for (int i1 = 0; i1 < sequences.size(); i1++) {
				for (int i2 = i1+1; i2 < sequences.size(); i2++) {
					var seq1 = sequences.get(i1);
					var seq2 = sequences.get(i2);

					BiFunction<Double, Double, Double> relationOperation = switch (lengthRelation) {
						case DIFFERENCE -> (a, b) -> a - b;
						case QUOTIENT -> (a, b) -> a / b;
					};
					accumulator.get(category).put(
						new Pair<>(seq1, seq2),
						new Pair<>(
							relationOperation.apply((double)Utils.getLength(seq1), (double)Utils.getLength(seq2)),
							(double)seq1.getDistance(seq2)
					));
		}}}

		return Map.copyOf(accumulator);
	}

	/**
	 * Prints the relation of the lengths and the distance for each pair of sequences in each category.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 */
	public static void pairwiseCorrelationPrint(ObjectMgmt mwMgmt, Utils.LengthRelation lengthRelation) {
		var pairwiseCorrelationMap = pairwiseCorrelation(mwMgmt, lengthRelation);

		for (var entry1 : pairwiseCorrelationMap.entrySet()) {
			var category = entry1.getKey();
			System.out.println("\nCategory: " + category.index);
			for (var entry2 : entry1.getValue().entrySet()) {
				var sequences = entry2.getKey();
				var results = entry2.getValue();

				System.out.println(
					"(" + sequences.first().getLocatorURI() + ", "
						+ sequences.second().getLocatorURI() + "): "
						+ results
				);
	}}}

	/**
	 * Creates an XY graph of the length relation and the distance for each pair of sequences in the category for each category.
	 * That means that the points of each category are shown in their own graph, separated from the points of other categories.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @param graphsPath path of the directory
	 */
	public static void pairwiseCorrelationShow(ObjectMgmt mwMgmt, Utils.LengthRelation lengthRelation, File graphsPath) {
		if (!graphsPath.isDirectory()) {
			throw new IllegalArgumentException(graphsPath.getPath() + " isn't a directory.");
		}

		var pairwiseCorrelationMap = pairwiseCorrelation(mwMgmt, lengthRelation);

		for (var entry : pairwiseCorrelationMap.entrySet()) {
			var dataset = new XYSeriesCollection();

			var category = entry.getKey();
			var values = entry.getValue();
			var dataSeries = new XYSeries(category.index);
			values.values().forEach(v -> dataSeries.add(
				(double)switch (lengthRelation) {
					case QUOTIENT -> Math.log(v.first()) / Math.log(2);
					default -> v.first();
				},
				v.second()
			));
			dataset.addSeries(dataSeries);

			var chart = ChartFactory.createScatterPlot(
				"The length relation and the distance of each two sequences in category " + category.index,
				"the " + switch(lengthRelation) {
					case DIFFERENCE -> "difference";
					case QUOTIENT -> "binary logarithm of the quotient";
				} +" of the lengths",
				"the distance",
				dataset
			);

			var graphFile = new File(graphsPath, "graph" + category.index + ".png");
			try {
				ChartUtils.saveChartAsPNG(graphFile, chart, 2048, 1024);
			} catch (IOException e) {
				System.err.println("An I/O error occurred in the writing of the chart to " + graphFile.getPath());
	}}}
}
